/*********************************************************************
 * webcam_server                                                     *
 *                                                                   *
 * (c) 2002 Donn Morrison donn@donn.dyndns.org                       *
 *                                                                   *
 * code used from Gerd Knorr's xawtv (libng)                         *
 * - and -                                                           *
 * Cory Lueninghoener's gqcam                                        *
 *                                                                   *
 *    waits for connections from a viewer and sends                  *
 *    jpeg encoded captures as a live video feed                     *
 *                                                                   *
 *********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/timeb.h>
#include <sys/mman.h>
#include <semaphore.h>
#include <netinet/in.h>
#include <unistd.h>
#include <linux/videodev.h>

#include "webcam_server.h"
#include "imgqueue.h"
#include "filters.h"
#include "camera.h"

#define MAX_BAD_GRABS		10	/* max bad image captures before exit */
#define CAPTURE_PAUSE		75000	/*  pause for 75ms between captures */

/*-----------------------------------------------------------
  grab_thread
  - runs only when clients are connected or when test mode
    enabled
  - captures an image, converts it to jpeg, pushes it on 
    queue, loops
  -----------------------------------------------------------*/
int grab_thread(struct caminfo *cam)
{
	char *jpeg_data = NULL;
	struct imagequeue *new;
	int len, width, height;
	u_long frames=0;
	time_t t0;
	time_t t1;
	struct timeb start_time;
	struct timeb stop_time;
	unsigned int capture_time;

	time(&t0);
	while(cam->grab_thread_alive)
	{
		if(cam->s.bad_grabs == MAX_BAD_GRABS)
			break;
	
		if(!cam->o.test_fps)
		{
			/* wait for a connection */
			sem_wait(&cam->sem_con);
			sem_post(&cam->sem_con);
		}

		ftime(&start_time);
		
		/* fetch an image */
		cam->raw_image = (char *)get_cam_image(cam, &len);
		if(cam->raw_image == NULL)
		{
#if DEBUG
			fprintf(stderr,"bad image grab (%d of %d)\n", cam->s.bad_grabs, MAX_BAD_GRABS);
			usleep(100000);
			cam->s.bad_grabs++;
#endif
			continue;
		}
		cam->s.bad_grabs = 0;
		
		time(&t1);
		if(cam->o.test_fps)
		{
			float fps;
			/* test mode only */
			frames++;
			fps = (float)frames/((t1 - t0) ? t1-t0 : 1.0);
			printf("read a frame: %d bytes, fps = %.3f, frames = %ld\n", len, fps, frames);
			continue;
		}

		/* swap rgb? */
		if(cam->o.swap_rgb)
			swap_rgb24(cam->raw_image, 
				   cam->o.width * cam->o.height);
		
		/* adjust gamma? */
		if(cam->o.gamma)
			adjust_gamma(cam->raw_image,
				     cam->o.width * cam->o.height * 3, cam->o.gamma);

		/* flip horiz/vert */
		if(cam->o.flip_horiz)
			fliph(cam->raw_image, cam->o.width, cam->o.height);
		if(cam->o.flip_vert)
			flipv(cam->raw_image, cam->o.width, cam->o.height);

		/* save width and height before potential rotate */
		width = cam->o.width;
		height = cam->o.height;

		/* rotate image? */
		if(cam->o.rotate > 0)
			rotate_image(cam->raw_image, &width, &height,
				    cam->o.rotate, 0, 0, cam->o.height, cam->o.width);

		/* add the caption? */
		if(cam->o.caption)
			add_text(cam, 
				 cam->raw_image, 
				 width, 
				 height);
		
		/* allocate some storage for the new image */
		new = (struct imagequeue *)malloc(sizeof(struct imagequeue));
		if(!new)
		{
			perror("malloc");
			clean(cam);
			continue;
		}
		
		/* compress the image in jpeg format */
		len = raw_to_jpeg(cam->raw_image,
				  &jpeg_data, 
				  width, 
				  height, 
				  cam->o.jpeg_quality);

		pthread_mutex_lock(&cam->lock_queue);
		pthread_mutex_lock(&cam->lock_id);

		/* set fields */
		new->id = ++(cam->s.id_newest_img);
		new->len = len;
		new->jpeg_data = jpeg_data;
		new->ref = 0;
		new->next = NULL;
		pthread_mutex_init(&new->lock_ref, NULL);
	
		/* put it on the queue */
		push(cam, new);
		
		pthread_mutex_unlock(&cam->lock_id);

		/* calculate optimum capture rate to save CPU time */
		ftime(&stop_time);
		capture_time = (stop_time.time - start_time.time);
		if (capture_time == 0)
		{
			capture_time = stop_time.millitm - start_time.millitm;
		}
		else
		{
			capture_time *= 1000;
			capture_time += (stop_time.millitm - start_time.millitm);
		}
		capture_time *= 100;
		if (capture_time < CAPTURE_PAUSE)
			usleep(CAPTURE_PAUSE - capture_time);

		/* clear unused images from the queue */
		len = clean(cam);
		pthread_mutex_unlock(&cam->lock_queue);
	}
	
	fprintf(stderr,"grab thread exiting...\n");

	munmap(cam->mmap, cam->vid_mbuf.size);

	exit(1);
}
